@@ -5,36 +5,44 @@
     Flask-Security decorators module
 
     :copyright: (c) 2012-2019 by Matt Wright.
-    :copyright: (c) 2019-2021 by J. Christopher Wagner (jwag).
+    :copyright: (c) 2019-2024 by J. Christopher Wagner (jwag).
     :license: MIT, see LICENSE for more details.
 """
 
+from __future__ import annotations
+
 from collections import namedtuple
 import datetime
 from functools import wraps
 import typing as t
 
-from flask import Response, _request_ctx_stack, abort, current_app, g, redirect, request
+from flask import Response, abort, current_app, g, redirect, request
 from flask_login import current_user, login_required  # noqa: F401
 from flask_principal import Identity, Permission, RoleNeed, identity_changed
 from flask_wtf.csrf import CSRFError
 from werkzeug.local import LocalProxy
-from werkzeug.routing import BuildError
 
 from .proxies import _security, DecoratedView
+from .signals import user_unauthenticated
 from .utils import (
     FsPermNeed,
-    config_value,
+    config_value as cv,
     do_flash,
-    find_user,
     get_message,
     get_url,
+    is_user_authenticated,
+    lookup_identity,
     check_and_update_authn_fresh,
     json_error_response,
     set_request_attr,
+    simplify_url,
+    get_request_attr,
     url_for_security,
 )
 
+if t.TYPE_CHECKING:  # pragma: no cover
+    from flask.typing import ResponseValue
+
 # Convenient references
 _csrf = LocalProxy(lambda: current_app.extensions["csrf"])
 
@@ -57,35 +65,38 @@ def _get_unauthenticated_response(text=None, headers=None):
     return Response(text, 401, headers)
 
 
-def _get_unauthorized_response(text=None, headers=None):  # pragma: no cover
-    # People called this - even though it isn't public - no harm in keeping it.
-    return _get_unauthenticated_response(text, headers)
-
-
-def default_unauthn_handler(mechanisms, headers=None):
+def default_unauthn_handler(mechanisms=None, headers=None):
     """Default callback for failures to authenticate
 
     If caller wants JSON - return 401.
     If caller wants BasicAuth - return 401 (the WWW-Authenticate header is set).
     Otherwise - assume caller is html and redirect if possible to a login view.
-    We let Flask-Login handle this.
-
     """
+    user_unauthenticated.send(
+        current_app._get_current_object(),  # type: ignore[attr-defined]
+        _async_wrapper=current_app.ensure_sync,
+    )
     headers = headers or {}
-    msg = get_message("UNAUTHENTICATED")[0]
+    m, c = get_message("UNAUTHENTICATED")
 
-    if config_value("BACKWARDS_COMPAT_UNAUTHN"):
+    if cv("BACKWARDS_COMPAT_UNAUTHN"):
         return _get_unauthenticated_response(headers=headers)
     if _security._want_json(request):
-        payload = json_error_response(errors=msg)
+        payload = json_error_response(errors=m)
         return _security._render_json(payload, 401, headers, None)
 
     # Basic-Auth is often used to provide a browser based login form and then the
     # browser will always add the BasicAuth credentials. For that to work we need to
     # return 401 and not redirect to our login view.
     if "WWW-Authenticate" in headers:
-        return Response(msg, 401, headers)
-    return _security.login_manager.unauthorized()
+        return Response(m, 401, headers)
+
+    do_flash(m, c)
+    # Simplify the original URL to be relative (if possible) and set as 'next' parameter
+    login_url = url_for_security("login", _external=True)
+    next_url = simplify_url(login_url, request.url)
+    redirect_url = url_for_security("login", next=next_url)
+    return redirect(redirect_url)
 
 
 def default_reauthn_handler(within, grace):
@@ -98,90 +109,103 @@ def default_reauthn_handler(within, grace):
     m, c = get_message("REAUTHENTICATION_REQUIRED")
 
     if _security._want_json(request):
-        is_us = config_value("UNIFIED_SIGNIN")
+        from .webauthn import has_webauthn
+
+        is_us = cv("UNIFIED_SIGNIN")
         payload = json_error_response(errors=m)
         payload["reauth_required"] = True
         payload["unified_signin_enabled"] = is_us
+        payload["has_webauthn_verify_credential"] = has_webauthn(
+            current_user, cv("WAN_ALLOW_AS_VERIFY")
+        )
         return _security._render_json(payload, 401, None, None)
 
-    view = "us_verify" if config_value("UNIFIED_SIGNIN") else "verify"
+    view = "us_verify" if cv("UNIFIED_SIGNIN") else "verify"
     do_flash(m, c)
-    redirect_url = url_for_security(view, next=request.url)
+    # Simplify the original URL to be relative (if possible) and set as 'next' parameter
+    view_url = url_for_security(view, _external=True)
+    next_url = simplify_url(view_url, request.url)
+    redirect_url = url_for_security(view, next=next_url)
     return redirect(redirect_url)
 
 
-def default_unauthz_handler(func, params):
+def default_unauthz_handler(func_name, params):
     unauthz_message, unauthz_message_type = get_message("UNAUTHORIZED")
     if _security._want_json(request):
         payload = json_error_response(errors=unauthz_message)
         return _security._render_json(payload, 403, None, None)
-    view = config_value("UNAUTHORIZED_VIEW")
-    if view:
+    if view := cv("UNAUTHORIZED_VIEW"):
         if callable(view):
-            view = view()
+            if not (redirect_to := view()):
+                abort(403)
         else:
-            try:
-                view = get_url(view)
-            except BuildError:
-                view = None
+            redirect_to = get_url(view)
         do_flash(unauthz_message, unauthz_message_type)
-        redirect_to = "/"
-        if request.referrer and not request.referrer.split("?")[0].endswith(
-            request.path
-        ):
-            redirect_to = request.referrer
-
-        return redirect(view or redirect_to)
+        return redirect(redirect_to)
     abort(403)
 
 
 def _check_token():
-    # N.B. this isn't great Flask-Login 0.5.0 made this protected
-    # Issue https://github.com/maxcountryman/flask-login/issues/471
-    # was filed to restore public access. We want to call this via
-    # login_manager in case someone has overridden the login_manager which we
-    # allow.
-    if hasattr(_security.login_manager, "request_callback"):
-        # Pre 0.5.0
-        user = _security.login_manager.request_callback(request)
-    else:
-        user = _security.login_manager._request_callback(request)
-
-    if user and user.is_authenticated:
-        app = current_app._get_current_object()
-        _request_ctx_stack.top.user = user
-        identity_changed.send(app, identity=Identity(user.fs_uniquifier))
+    user = _security.login_manager.request_callback(request)
+    if is_user_authenticated(user):
+        identity_changed.send(
+            current_app._get_current_object(),
+            _async_wrapper=current_app.ensure_sync,
+            identity=Identity(user.fs_uniquifier),
+        )
         return True
 
     return False
 
 
+def _check_session():
+    """
+    Note that flask_login will have already run _load_user (due to someone referencing
+    current_user proxy). This will have called our _user_loader or _request_loader
+    already.
+    This method needs to determine whether the authenticated user was authenticated
+    due to _user_loader or _request_loader and return True for the former.
+    This routine makes sure that if an endpoint is decorated with just allowing
+    'session' that token authentication won't return True (even though the user
+    is in fact correctly authenticated). There are certainly endpoints that need to
+    be web browser/session only (often those that in fact return tokens!).
+    """
+    if not is_user_authenticated(current_user):
+        return False
+    if get_request_attr("fs_authn_via") != "session":
+        return False
+    return True
+
+
 def _check_http_auth():
     auth = request.authorization or BasicAuth(username=None, password=None)
     if not auth.username:
         return False
-    user = find_user(auth.username)
+    user = lookup_identity(auth.username)
     if user and not user.active:
         return False
 
     if user and user.verify_and_update_password(auth.password):
         _security.datastore.commit()
-        app = current_app._get_current_object()
-        _request_ctx_stack.top.user = user
-        identity_changed.send(app, identity=Identity(user.fs_uniquifier))
+        _security.login_manager._update_request_context_with_user(user)
+        identity_changed.send(
+            current_app._get_current_object(),
+            _async_wrapper=current_app.ensure_sync,
+            identity=Identity(user.fs_uniquifier),
+        )
         return True
 
     return False
 
 
-def handle_csrf(method: t.Optional[str]) -> None:
+def handle_csrf(method: str, json_response: bool = False) -> ResponseValue | None:
     """Invoke CSRF protection based on authentication method.
 
     Usually this is called as part of a decorator, but if that isn't
     appropriate, endpoint code can call this directly.
 
     If CSRF protection is appropriate, this will call flask_wtf::protect() which
-    will raise a ValidationError on CSRF failure.
+    will raise a CSRFError(BadRequest) on CSRF failure.
 
     This routine does nothing if any of these are true:
 
@@ -191,24 +215,44 @@ def handle_csrf(method: t.Optional[str]) -> None:
 
         #) csrfProtect already checked and accepted the token
 
-    If the passed in method is not in *SECURITY_CSRF_PROTECT_MECHANISMS* then not only
-    will no CSRF code be run, but a flag in the current context ``fs_ignore_csrf``
-    will be set so that downstream code knows to ignore any CSRF checks.
+    This means in the default config - CSRF is done as part of form validation
+    not here. Only if the application calls CSRFProtect(app) will this method
+    do anything. Furthermore - since this is called PRIOR to form instantiation
+    if the request is JSON - it MUST send the csrf_token as a header.
+
+    If the passed in method is not in
+    :py:data:`SECURITY_CSRF_PROTECT_MECHANISMS` then in addition to
+    no CSRF code being run, the flask_wtf request global 'csrf_valid' will be set
+    so that downstream code knows to ignore any CSRF checks.
+
+    Returns None if all ok, returns a Response with JSON error if request
+    wanted JSON - else re-raises the CSRFError exception.
 
     .. versionadded:: 3.3.0
+
+    .. versionchanged:: 5.4.3
+        Use flask_wtf request global 'csrf_valid' instead of our own to handle
+        application forms that aren't derived from our forms.
     """
     if (
         not current_app.config.get("WTF_CSRF_ENABLED", False)
         or not current_app.extensions.get("csrf", None)
         or g.get("csrf_valid", False)
     ):
-        return
+        return None
 
-    if config_value("CSRF_PROTECT_MECHANISMS"):
-        if method in config_value("CSRF_PROTECT_MECHANISMS"):
-            _csrf.protect()  # type: ignore
-        else:
-            _request_ctx_stack.top.fs_ignore_csrf = True
+    if cv("CSRF_PROTECT_MECHANISMS"):
+        if method in cv("CSRF_PROTECT_MECHANISMS"):
+            try:
+                _csrf.protect()  # type: ignore
+            except CSRFError as e:
+                if json_response:
+                    payload = json_error_response(errors=e.description)
+                    return _security._render_json(payload, 400, None, None)
+                raise
+            return None
+    set_request_attr("csrf_valid", True)  # flask_wtf global
+    return None
 
 
 def http_auth_required(realm: t.Any) -> DecoratedView:
@@ -226,15 +270,14 @@ def http_auth_required(realm: t.Any) -> DecoratedView:
         @wraps(fn)
         def wrapper(*args, **kwargs):
             if _check_http_auth():
-                handle_csrf("basic")
+                eresponse = handle_csrf("basic", _security._want_json(request))
+                if eresponse:
+                    return eresponse
                 set_request_attr("fs_authn_via", "basic")
-                return fn(*args, **kwargs)
-            if _security._unauthorized_callback:
-                return _security._unauthorized_callback()
-            else:
-                r = _security.default_http_auth_realm if callable(realm) else realm
-                h = {"WWW-Authenticate": 'Basic realm="%s"' % r}
-                return _security._unauthn_handler(["basic"], headers=h)
+                return current_app.ensure_sync(fn)(*args, **kwargs)
+            r = cv("DEFAULT_HTTP_AUTH_REALM") if callable(realm) else realm
+            h = {"WWW-Authenticate": f'Basic realm="{r}"'}
+            return _security._unauthn_handler(["basic"], headers=h)
 
         return wrapper
 
@@ -247,8 +290,8 @@ def auth_token_required(fn: DecoratedView) -> DecoratedView:
     """Decorator that protects endpoints using token authentication. The token
     should be added to the request by the client by using a query string
     variable with a name equal to the configuration value of
-    *SECURITY_TOKEN_AUTHENTICATION_KEY* or in a request header named that of
-    the configuration value of *SECURITY_TOKEN_AUTHENTICATION_HEADER*
+    :py:data:`SECURITY_TOKEN_AUTHENTICATION_KEY` or in a request header named that of
+    the configuration value of :py:data:`SECURITY_TOKEN_AUTHENTICATION_HEADER`
 
     Once authenticated, if so configured, CSRF protection will be tested.
     """
@@ -256,24 +299,23 @@ def auth_token_required(fn: DecoratedView) -> DecoratedView:
     @wraps(fn)
     def decorated(*args, **kwargs):
         if _check_token():
-            handle_csrf("token")
+            eresponse = handle_csrf("token", _security._want_json(request))
+            if eresponse:
+                return eresponse
             set_request_attr("fs_authn_via", "token")
-            return fn(*args, **kwargs)
-        if _security._unauthorized_callback:
-            return _security._unauthorized_callback()
-        else:
-            return _security._unauthn_handler(["token"])
+            return current_app.ensure_sync(fn)(*args, **kwargs)
+        return _security._unauthn_handler(["token"])
 
     return t.cast(DecoratedView, decorated)
 
 
 def auth_required(
-    *auth_methods: t.Union[str, t.Callable[[], t.List[str]], None],
-    within: t.Union[int, float, t.Callable[[], datetime.timedelta]] = -1,
-    grace: t.Optional[t.Union[int, float, t.Callable[[], datetime.timedelta]]] = None
+    *auth_methods: str | t.Callable[[], list[str]] | None,
+    within: int | float | t.Callable[[], datetime.timedelta] = -1,
+    grace: int | float | t.Callable[[], datetime.timedelta] | None = None,
 ) -> DecoratedView:
     """
-    Decorator that protects endpoints through multiple mechanisms
+    Decorator that protects endpoints through multiple mechanisms.
     Example::
 
         @app.route('/dashboard')
@@ -290,9 +332,7 @@ def auth_required(
         timedelta.total_seconds() is used for the calculations:
 
             - If > 0, then the caller must have authenticated within the time specified
-              (as measured using the session cookie).
-            - If 0 and not within the grace period (see below) the caller will
-              always be redirected to re-authenticate.
+              (as measured using the session cookie or authentication token).
             - If < 0 (the default) no freshness check is performed.
 
         Note that Basic Auth, by definition, is always 'fresh' and will never result in
@@ -309,8 +349,7 @@ def auth_required(
     The first mechanism that succeeds is used, following that, depending on
     configuration, CSRF protection will be tested.
 
-    On authentication failure `.Security.unauthorized_callback` (deprecated)
-    or :meth:`.Security.unauthn_handler` will be called.
+    On authentication failure :meth:`.Security.unauthn_handler` will be called.
 
     As a side effect, upon successful authentication, the request global
      ``fs_authn_via`` will be set to the method ("basic", "token", "session")
@@ -338,7 +377,7 @@ def auth_required(
 
     login_mechanisms = {
         "token": lambda: _check_token(),
-        "session": lambda: current_user.is_authenticated,
+        "session": lambda: _check_session(),
         "basic": lambda: _check_http_auth(),
     }
     mechanisms_order = ["token", "session", "basic"]
@@ -353,7 +392,7 @@ def auth_required(
             else:
                 within = datetime.timedelta(minutes=within)
             if grace is None:
-                grace = config_value("FRESHNESS_GRACE_PERIOD")
+                grace = cv("FRESHNESS_GRACE_PERIOD")
             elif callable(grace):
                 grace = grace()
             else:
@@ -371,8 +410,8 @@ def auth_required(
 
             h = {}
             if "basic" in ams:
-                r = _security.default_http_auth_realm
-                h["WWW-Authenticate"] = 'Basic realm="%s"' % r
+                r = cv("DEFAULT_HTTP_AUTH_REALM")
+                h["WWW-Authenticate"] = f'Basic realm="{r}"'
             mechanisms = [
                 (method, login_mechanisms.get(method))
                 for method in mechanisms_order
@@ -381,17 +420,14 @@ def auth_required(
             for method, mechanism in mechanisms:
                 if mechanism and mechanism():
                     # successfully authenticated. Basic auth is by definition 'fresh'.
-                    # Note that using token auth is ok - but caller still has to pass
-                    # in a session cookie...
+                    # If 'within' is set - check for freshness of authentication.
                     if not check_and_update_authn_fresh(within, grace, method):
                         return _security._reauthn_handler(within, grace)
-                    handle_csrf(method)
+                    if eresponse := handle_csrf(method, _security._want_json(request)):
+                        return eresponse
                     set_request_attr("fs_authn_via", method)
-                    return fn(*args, **kwargs)
-            if _security._unauthorized_callback:
-                return _security._unauthorized_callback()
-            else:
-                return _security._unauthn_handler(ams, headers=h)
+                    return current_app.ensure_sync(fn)(*args, **kwargs)
+            return _security._unauthn_handler(ams, headers=h)
 
         return decorated_view
 
@@ -412,42 +448,42 @@ def unauth_csrf(
 
     This decorator does nothing if *WTF_CSRF_ENABLED* == **False**.
 
-    This decorator will always require CSRF if the caller is authenticated.
+    This decorator does nothing if the caller is authenticated.
 
     This decorator will suppress CSRF if caller isn't authenticated and has set the
-    *SECURITY_CSRF_IGNORE_UNAUTH_ENDPOINTS* config variable.
-
-    :param fall_through: if set to True, then if CSRF fails here - simply keep going.
-        This is appropriate if underlying view is form based and once the form is
-        instantiated, the csrf_token will be available.
-        Note that this can mask some errors such as 'The CSRF session token is missing.'
-        meaning that the caller didn't send a session cookie and instead the caller
-        might get a 'The CSRF token is missing.' error.
+    :py:data:`SECURITY_CSRF_IGNORE_UNAUTH_ENDPOINTS` config variable to **True**.
 
     .. versionadded:: 3.3.0
+
+    .. versionchanged:: 5.4.3
+        The fall_through parameter is now ignored.
+        Add code to properly handle JSON errors.
     """
 
     def wrapper(fn):
         @wraps(fn)
         def decorated(*args, **kwargs):
-            if not current_app.config.get(
-                "WTF_CSRF_ENABLED", False
-            ) or not current_app.extensions.get("csrf", None):
-                return fn(*args, **kwargs)
-
             if (
-                config_value("CSRF_IGNORE_UNAUTH_ENDPOINTS")
-                and not current_user.is_authenticated
+                not current_app.config.get("WTF_CSRF_ENABLED", False)
+                or not current_app.extensions.get("csrf", None)
+                or g.get("csrf_valid", False)
             ):
-                _request_ctx_stack.top.fs_ignore_csrf = True
+                return current_app.ensure_sync(fn)(*args, **kwargs)
+
+            if cv("CSRF_IGNORE_UNAUTH_ENDPOINTS") and not is_user_authenticated(
+                current_user
+            ):
+                set_request_attr("csrf_valid", True)
             else:
                 try:
                     _csrf.protect()
-                except CSRFError:
-                    if not fall_through:
-                        raise
+                except CSRFError as e:
+                    if _security._want_json(request):
+                        payload = json_error_response(errors=e.description)
+                        return _security._render_json(payload, 400, None, None)
+                    raise
 
-            return fn(*args, **kwargs)
+            return current_app.ensure_sync(fn)(*args, **kwargs)
 
         return decorated
 
@@ -475,11 +511,10 @@ def roles_required(*roles: str) -> DecoratedView:
             perms = [Permission(RoleNeed(role)) for role in roles]
             for perm in perms:
                 if not perm.can():
-                    if _security._unauthorized_callback:
-                        # Backwards compat - deprecated
-                        return _security._unauthorized_callback()
-                    return _security._unauthz_handler(roles_required, list(roles))
-            return fn(*args, **kwargs)
+                    return _security._unauthz_handler(
+                        roles_required.__name__, list(roles)
+                    )
+            return current_app.ensure_sync(fn)(*args, **kwargs)
 
         return decorated_view
 
@@ -506,11 +541,8 @@ def roles_accepted(*roles: str) -> DecoratedView:
         def decorated_view(*args, **kwargs):
             perm = Permission(*(RoleNeed(role) for role in roles))
             if perm.can():
-                return fn(*args, **kwargs)
-            if _security._unauthorized_callback:
-                # Backwards compat - deprecated
-                return _security._unauthorized_callback()
-            return _security._unauthz_handler(roles_accepted, list(roles))
+                return current_app.ensure_sync(fn)(*args, **kwargs)
+            return _security._unauthz_handler(roles_accepted.__name__, list(roles))
 
         return decorated_view
 
@@ -542,14 +574,11 @@ def permissions_required(*fsperms: str) -> DecoratedView:
             perms = [Permission(FsPermNeed(fsperm)) for fsperm in fsperms]
             for perm in perms:
                 if not perm.can():
-                    if _security._unauthorized_callback:
-                        # Backwards compat - deprecated
-                        return _security._unauthorized_callback()
                     return _security._unauthz_handler(
-                        permissions_required, list(fsperms)
+                        permissions_required.__name__, list(fsperms)
                     )
 
-            return fn(*args, **kwargs)
+            return current_app.ensure_sync(fn)(*args, **kwargs)
 
         return decorated_view
 
@@ -580,11 +609,10 @@ def permissions_accepted(*fsperms: str) -> DecoratedView:
         def decorated_view(*args, **kwargs):
             perm = Permission(*(FsPermNeed(fsperm) for fsperm in fsperms))
             if perm.can():
-                return fn(*args, **kwargs)
-            if _security._unauthorized_callback:
-                # Backwards compat - deprecated
-                return _security._unauthorized_callback()
-            return _security._unauthz_handler(permissions_accepted, list(fsperms))
+                return current_app.ensure_sync(fn)(*args, **kwargs)
+            return _security._unauthz_handler(
+                permissions_accepted.__name__, list(fsperms)
+            )
 
         return decorated_view
 
@@ -594,7 +622,7 @@ def permissions_accepted(*fsperms: str) -> DecoratedView:
 def anonymous_user_required(f: DecoratedView) -> DecoratedView:
     """Decorator which requires that caller NOT be logged in.
     If a logged in user accesses an endpoint protected with this decorator
-    they will be redirected to the *SECURITY_POST_LOGIN_VIEW*.
+    they will be redirected to the :py:data:`SECURITY_POST_LOGIN_VIEW`.
     If the caller requests a JSON response, a 400 will be returned.
 
     .. versionchanged:: 3.3.0
@@ -603,14 +631,14 @@ def anonymous_user_required(f: DecoratedView) -> DecoratedView:
 
     @wraps(f)
     def wrapper(*args, **kwargs):
-        if current_user.is_authenticated:
+        if is_user_authenticated(current_user):
             if _security._want_json(request):
                 payload = json_error_response(
                     errors=get_message("ANONYMOUS_USER_REQUIRED")[0]
                 )
                 return _security._render_json(payload, 400, None, None)
             else:
-                return redirect(get_url(_security.post_login_view))
-        return f(*args, **kwargs)
+                return redirect(get_url(cv("POST_LOGIN_VIEW")))
+        return current_app.ensure_sync(f)(*args, **kwargs)
 
     return t.cast(DecoratedView, wrapper)
